// -*- C++ -*-
/*!
 * @file CORBA_SeqUtil.h
 * @brief CORBA sequence utility template functions
 * @date $Date: 2007-12-31 03:06:24 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006-2008
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: CORBA_SeqUtil.h 2404 2014-02-06 02:15:22Z n-ando $
 *
 */

#ifndef CORBA_SEQUTIL_H
#define CORBA_SEQUTIL_H

#include <rtm/RTC.h>
#include <rtm/Manager.h>

/*!
 * @if jp
 * @namespace CORBA_SeqUtil
 *
 * @brief CORBA sequence wp[ev[g֐
 *
 * CORBA sequence ɑ΂Ĉȉ̃[eBeBev[g֐񋟂B
 * ̓XbhZ[tł͂Ȃ̂ŁAXbhZ[tɑ삵ꍇ́A
 * ΏۂƂȂV[PXlK؂mutexŕی삷KvB
 *
 * - for_each()
 * - find()
 * - push_back()
 * - insert()
 * - front()
 * - back()
 * - erase()
 * - clear()
 *
 * @else
 *
 * @namespace CORBA_SeqUtil
 *
 * @brief CORBA sequence helper template functions
 *
 * This group provides the following utility function to CORBA sequence.
 * Since these functions are not thread-safe operations,
 * if the sequence would be operated in thread-safe,
 * the value should be protected by mutex properly.
 *
 * - for_each()
 * - find()
 * - push_back()
 * - insert()
 * - front()
 * - back()
 * - erase()
 * - clear()
 *
 * @endif
 */
namespace CORBA_SeqUtil
{
  /*!
   * @if jp
   *
   * @brief CORBA sequence ɑ΂ functor Kp
   *
   * CORBA sequence SĂ̗vfɑ΂āA^ꂽ functor KpB
   * functor  void functor(CORBA sequence ̗vf) ̌`ƂKvB
   *
   * @param seq Functor Kp CORBA sequence
   * @param f CORBA sequence ̗vf Functor
   *
   * @return SĂ̗vf Functor
   *
   * @since 0.4.0
   *
   * @else
   *
   * @brief Apply the functor to all CORBA sequence elements
   *
   * Apply the given functor to the given CORBA sequence.
   * functor should be void functor(CORBA sequence element).
   *
   * @param seq CORBA sequence to be applied the functor
   * @param f A functor to process CORBA sequence elements
   *
   * @return Functor that processed all CORBA sequence elements
   *
   * @endif
   */
  template <class CorbaSequence, class Functor>
  Functor for_each(CorbaSequence& seq, Functor f)
  {
    CORBA::ULong len;
    len = seq.length();
    for (CORBA::ULong i = 0; i < len; ++i)
      {
        f(seq[i]);
      }
    return f;
  }

  /*!
   * @if jp
   * @brief CORBA sequence ̒ functor ɓKvf̃CfbNXԂ
   *
   * CORBA sequence SĂ̗vfɑ΂āA^ꂽ functor KpA
   * functor  true Ԃ悤̃CfbNXԂB
   * functor  bool functor(const CORBA sequence ̗vf) ̌`ƂA
   * Kvfɑ΂ true ԂKvB
   *
   * @param seq Functor Kp CORBA sequence
   * @param f CORBA sequence vf Functor
   *
   * @return Functor ɓKvf̃CfbNXB
   *         ȂƂ -1 ԂB
   *
   * @else
   *
   * @brief Return the index of CORBA sequence element that functor matches
   *
   * This operation applies the given functor to the given CORBA sequence,
   * and returns the index of the sequence element that the functor matches.
   * The functor should be bool functor(const CORBA sequence element) type,
   * and it would return true, if the element matched the functor.
   *
   * @param seq CORBA sequence to be applied the functor
   * @param f A functor to process CORBA sequence elements
   *
   * @return The index of the element that functor matches.
   *         If no element found, it would return -1.
   *
   * @endif
   */
  template <class CorbaSequence, class Functor>
  CORBA::Long find(const CorbaSequence& seq, Functor f)
  {
    CORBA::ULong len(seq.length());
    for (CORBA::ULong i = 0; i < len; ++i)
      {
        if (f(seq[i])) { return (CORBA::Long)i; }
      }
    return -1;
  }

  /*!
   * @if jp
   * @brief CORBA sequence ̍Ōɗvfǉ
   *
   * CORBA sequence ̍Ōɗ^ꂽvfǉB
   * CORBA sequence ͎̒IɊgB
   *
   * @param seq vfǉ CORBA sequence
   * @param elem ǉvf
   *
   * @else
   *
   * @brief Push the new element back to the CORBA sequence
   *
   * Add the given element to the last of CORBA sequence.
   * The length of the CORBA sequence will be expanded automatically.
   *
   * @param seq CORBA sequence to be added a new element
   * @param elem The new element to be added to the CORBA sequence
   *
   * @endif
   */
  template <class CorbaSequence, class SequenceElement>
  void push_back(CorbaSequence& seq, SequenceElement elem)
  {
    CORBA::ULong len(seq.length());
    seq.length(len + 1);
    seq[len] = elem;
  }

  /*!
   * @if jp
   * @brief CORBA sequence }[W
   *
   * ^ꂽ CORBA sequence }[WB
   *
   * @param seq1 }[W CORBA sequence
   * @param seq2 }[W CORBA sequence
   *
   * @else
   *
   * @brief Merge the elements of the CORBA sequence
   *
   * Merge given CORBA sequences.
   *
   * @param seq1 merge target CORBA sequence
   * @param seq2 merge target CORBA sequence
   *
   * @endif
   */
  template <class CorbaSequence>
  void push_back_list(CorbaSequence& seq1, const CorbaSequence& seq2)
  {
    CORBA::ULong len1(seq1.length());
    CORBA::ULong len2(seq2.length());
    CORBA::ULong len(len1 + len2);
    seq1.length(len);

    for (CORBA::ULong i = 0; i < len2; ++i)
      {
        seq1[len1 + i] = seq2[i];
      }
  }

  /*!
   * @if jp
   * @brief CORBA sequence ɗvf}
   *
   * CORBA sequence  index ̈ʒuɗvfB
   * index  ^ꂽ@CORBA sequence ̍ő index 傫ꍇ
   * Ō̗vfƂĉB
   * CORBA sequence ͎̒IɊgB
   *
   * @param seq vfǉ CORBA sequence
   * @param elem ǉvf
   * @param index vfǉʒu
   *
   * @else
   *
   * @brief Insert the element to the CORBA sequence
   *
   * Insert a new element in the given position to the CORBA sequence.
   * If the given index is greater than the length of the sequence,
   * the given element is pushed back to the last of the sequence.
   * The length of the CORBA sequence will be expanded automatically.
   *
   * @param seq The CORBA sequence to be inserted a new element
   * @param elem The new element to be inserted the sequence
   * @param index The inserting position
   *
   * @endif
   */
  template <class CorbaSequence, class SequenceElement>
  void insert(CorbaSequence& seq, SequenceElement& elem, CORBA::ULong index)
  {
    CORBA::ULong len(seq.length());
    if (index > len)
      {
        push_back(seq, elem);
        return;
      }

    seq.length(len + 1);
    for (CORBA::ULong i = len; i > index; --i)
      {
        seq[i] = seq[i - 1];
      }
    seq[index] = elem;
  }

  /*!
   * @if jp
   * @brief CORBA sequence ̐擪vf擾
   *
   * CORBA sequence ̐擪vf擾B
   * seq[0] ƓB
   *
   * @param seq vf擾 CORBA sequence
   *
   * @return 擾vf
   *
   * @else
   *
   * @brief Get the front element of the CORBA sequence
   *
   * This operation returns seq[0].
   *
   * @param seq CORBA sequence which acquires an element
   *
   * @return An acquisition element
   *
   * @param seq The CORBA sequence to be get the element
   *
   * @endif
   */
  template <class CorbaSequence, class SequenceElement>
  SequenceElement& front(CorbaSequence& seq)
  {
    return seq[0];
  }

  /*!
   * @if jp
   * @brief CORBA sequence ̖vf擾
   *
   * CORBA sequence ̖vf擾B
   * seq[seq.length() - 1] ƓB
   *
   * @param seq vf擾 CORBA sequence
   *
   * @return 擾vf
   *
   * @else
   *
   * @brief Get the last element of the CORBA sequence
   *
   * This operation returns seq[seq.length() - 1].
   *
   * @param seq The CORBA sequence to be get the element
   *
   * @return An acquisition element
   *
   * @endif
   */
  template <class CorbaSequence, class SequenceElement>
  SequenceElement& back(CorbaSequence& seq)
  {
    return seq[seq.length() - 1];
  }

  /*!
   * @if jp
   * @brief CORBA sequence ̎w肳ꂽʒu̗vf폜
   *
   * w肳ꂽCfbNX̗vf폜B
   * 폜ꂽvf͋l߂Asequence ̒1B
   *
   * @param seq vf폜 CORBA sequence
   * @param index 폜vf̃CfbNX
   *
   * @else
   *
   * @brief Erase the element of the specified index
   *
   * This operation removes the element of the given index.
   * The other elements are closed up around the hole.
   *
   * @param seq The CORBA sequence to be get the element
   * @param index The index of the element to be removed
   *
   * @endif
   */
  template <class CorbaSequence>
  void erase(CorbaSequence& seq, CORBA::ULong index)
  {
    CORBA::ULong len(seq.length());
    if (index > len) return;

    for (CORBA::ULong i = index; i < len - 1; ++i)
      {
        seq[i] = seq[i + 1];
      }
    seq.length(len - 1);
  }

  /*!
   * @if jp
   *
   * @brief V[PX̗vfqɂč폜
   *
   * ̃Iy[V͏qƂė^ꂽ֐IuWFNg
   * ^̂ƂÃV[PX̗vf폜B
   *
   * @param seq vfΏۂ CORBA sequence
   * @param f 폜V[PX肷p
   *
   * @else
   *
   * @brief Remove an element of a sequence according to a predicate
   *
   * This operation removes the element from sequence when a condition of
   * a function object given as a predicate is True.
   *
   * @param seq target CORBA sequence
   * @param f predicate which decides a sequence to remove
   *
   * @endif
   */
  template <class CorbaSequence, class Functor>
  void erase_if(CorbaSequence& seq, Functor f)
  {
    CORBA::Long index;
    index = find(seq, f);
    if (index < 0) return;
    CORBA_SeqUtil::erase(seq, index);
  }

  /*!
   * @if jp
   * @brief CORBA sequence ̑Svf폜
   *
   * CORBA sequence ̑Svf폜B
   * seq.length(0) ƓB
   *
   * @else
   *
   * @brief Erase all the elements of the CORBA sequence
   *
   * same as seq.length(0).
   *
   * @endif
   */
  template <class CorbaSequence>
  void clear(CorbaSequence& seq)
  {
    seq.length(0);
  }
  // End of CORBA sequence helper functions


  template <class CorbaRefSequence>
  coil::vstring refToVstring(const CorbaRefSequence& objlist)
  {
    coil::vstring iorlist;
    CORBA::ORB_ptr orb = ::RTC::Manager::instance().getORB();

    for (CORBA::ULong i(0), len(objlist.length()); i < len; ++i)
      {
#ifndef ORB_IS_RTORB
        CORBA::String_var str_var = orb->object_to_string(objlist[i].in());
#else
        CORBA_Object obj = (objlist.cobj())->_buffer[i];
        CORBA::String_var str_var = orb->object_to_string2(obj);
#endif
        iorlist.push_back(str_var.in());
      }
    return iorlist;
  }
}; // namespace CORBA_SeqUtil
#endif // CORBA_SEQUTIL_H
